import tarfile
from enum import Enum
from importlib import resources
from pathlib import Path

import appdirs
from pydantic import DirectoryPath, validate_arguments

import bigearthnet_common


class ExampleResource(str, Enum):
    s1_example_bz2_tar = "BigEarthNet-S1-Example.tar.bz2"
    s2_example_bz2_tar = "BigEarthNet-S2-Example.tar.bz2"

    def __str__(self):
        return self.value


USER_DIR = Path(appdirs.user_data_dir("bigearthnet_common"))
USER_DIR.mkdir(exist_ok=True, parents=True)


@validate_arguments
def _extract_sentinel_resource(
    resource: ExampleResource,
    subfolder_name: str,
    target_root_dir: DirectoryPath = USER_DIR,
) -> Path:
    """
    Given a local `resource` (one of `ExampleResource`), extract the `subfolder_name` of the compressed
    tar file to the given `target_root_directory`.
    The output path will be `<target_root_dir>/<subfolder_name>`
    Where `subfolder_name` _must_ be a root-folder of the compressed archive!

    The archive will _always_ be extracted. The function applies no caching!
    """
    extracted_dir = target_root_dir / subfolder_name

    if not resources.is_resource(bigearthnet_common, resource):
        raise ValueError(
            f"{resource} resource is not available! This means that it was forgotten to be packaged."
        )
    with resources.path(bigearthnet_common, resource) as resource_path:
        tar_file = tarfile.open(resource_path)
        folder_dir = tar_file.getnames()[0]
        assert folder_dir == subfolder_name
        tar_file.extractall(target_root_dir)
    assert extracted_dir.exists()
    return extracted_dir


def get_s1_example_folder_path() -> Path:
    """Get the path to a tiny subset of BigEarthNet-S1"""
    resource = ExampleResource.s1_example_bz2_tar
    subfolder = "BigEarthNet-S1-Example"
    return _extract_sentinel_resource(resource, subfolder)


def get_s2_example_folder_path() -> Path:
    """Get the path to a tiny subset of BigEarthNet-S2"""
    resource = ExampleResource.s2_example_bz2_tar
    subfolder = "BigEarthNet-S2-Example"
    return _extract_sentinel_resource(resource, subfolder)


def get_s1_example_patch_path() -> Path:
    """Get the path to a single BigEarthNet-S1 patch."""
    s1_dir = get_s1_example_folder_path()
    return sorted(p for p in s1_dir.iterdir())[0]


def get_s2_example_patch_path() -> Path:
    """Get the path to a single BigEarthNet-S2 patch."""
    s2_dir = get_s2_example_folder_path()
    return sorted(p for p in s2_dir.iterdir())[0]



"""
BigEarthNet Base Functions:

A collection of common functions that are frequently used when working with BigEarthNet.
"""

import bz2
import csv
import functools
import json
import warnings
from datetime import datetime
from enum import Enum
from importlib import resources
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, Union

import appdirs
import dateutil
import fastcore.all as fc
import rich
import rich_click.typer as typer
from fastcore.basics import compose
from fastcore.dispatch import typedispatch
from pydantic import DirectoryPath, FilePath, validate_arguments

# from importlib_resources import files
from rich.table import Table

import bigearthnet_common
import bigearthnet_common.constants as ben_constants


class Resource(str, Enum):
    patches_with_cloud_and_snow = "patches_with_cloud_and_shadow.csv.bz2"
    patches_with_seasonal_snow = "patches_with_seasonal_snow.csv.bz2"
    patches_with_no_19_class_targets = "patches_with_no_19_class_targets.csv.bz2"
    s1_s2_mapping = "s1_s2_name_country_season.csv.bz2"
    country = "s1_s2_name_country_season.csv.bz2"
    season = "s1_s2_name_country_season.csv.bz2"
    test_csv = "test.csv.bz2"
    train_csv = "train.csv.bz2"
    val_csv = "val.csv.bz2"

    def __str__(self):
        return self.value


USER_DIR = Path(appdirs.user_data_dir("bigearthnet"))
USER_DIR.mkdir(exist_ok=True, parents=True)
PATCHES_WITH_SNOW_URL = (
    "http://bigearth.net/static/documents/patches_with_seasonal_snow.csv"
)
PATCHES_WITH_CLOUD_AND_SHADOW_URL = (
    "http://bigearth.net/static/documents/get_patches_with_cloud_and_shadow.csv"
)


def parse_datetime(inp: Union[str, datetime]) -> datetime:
    """
    Parses an input into a `datetime` object.
    Will try its best to infer the correct format from a string.
    If a `datetime` object is already provided it will be returned.
    Otherwise it will raise an error.
    """
    return _parse_datetime(inp)


@typedispatch
def _parse_datetime(acquisition_date: str) -> datetime:
    return compose(dateutil.parser.parse, _parse_datetime)(acquisition_date)


@typedispatch
def _parse_datetime(acquisition_date: datetime) -> datetime:
    return acquisition_date


@typedispatch
def _parse_datetime(acquisition_date: object) -> None:
    raise TypeError("Could not parse acquisition_date!")


@validate_arguments
def _read_json(
    json_fp: FilePath, expected_keys: Set, read_only_expected: bool = True
) -> Dict[str, str]:
    """
    Parse the json file given with the file path `json_fp`.
    The function checks if all of the `expected_keys` are present, which
    ensures that no keys have been accidentilly deleted (this has happend before).
    If `read_only_expected` is set, only the keys provided in `expected_keys` are read
    and returned.
    This prevents accidental processing of injected metadata.

    Args:
        json_fp (FilePath): Path to json file
        expected_keys (Set): Keys that are expected to be present in the json file
        read_only_expected (bool, optional): Read only the keys given in `expected_keys`. Defaults to True.

    Returns:
        [Dict[str, str]]: A dictionary of the keys.
    """
    try:
        complete_data = json.loads(json_fp.read_text())
    except json.JSONDecodeError:
        raise ValueError(f"Error trying to read json from: ", json_fp)

    missing_elements = expected_keys - complete_data.keys()
    if len(missing_elements) > 0:
        raise ValueError(f"{json_fp} is missing entries!", missing_elements)

    # ensure that the original values are loaded, as some users may customize the original json files
    if read_only_expected:
        return {k: v for k, v in complete_data.items() if k in expected_keys}
    return complete_data


def read_S1_json(json_fp: FilePath) -> Dict[str, str]:
    """
    A helper function that *safely* reads a BigEarthNet-S1 json file.
    It will ensure that all expected entries are present and only read those
    entries.
    This helps to avoid issues where the JSON files were accidentally modified
    or partially deleted.

    Note: This function will also silently fix a typo present in the `coordinates` key
    from version: S1_v1.0. A coordinates key is named `lly` and it should be `lry`.
    """
    data = _read_json(json_fp, ben_constants.BEN_S1_V1_0_JSON_KEYS)
    # Silently fix key error in S1
    if "lly" in data["coordinates"]:
        data["coordinates"]["lry"] = data["coordinates"].pop("lly")
    return data


def read_S2_json(json_fp: FilePath) -> Dict[str, str]:
    """
    A helper function that *safely* reads a BigEarthNet-S2 json file.
    It will ensure that all expected entries are present and only read those
    entries.
    This helps to avoid issues where the JSON files were accidentally modified
    or partially deleted.
    """
    return _read_json(json_fp, ben_constants.BEN_S2_V1_0_JSON_KEYS)


@validate_arguments
def get_s2_patch_directories(dir_path: DirectoryPath) -> List[Path]:
    """
    Will find all S2 patch directories in the provided `dir_path`.
    Only directories that strictly cohere to the naming convention will be returned.
    """
    return [
        p
        for p in dir_path.iterdir()
        if ben_constants.BEN_S2_RE.fullmatch(p.name) is not None
    ]


@validate_arguments
def get_s1_patch_directories(dir_path: DirectoryPath) -> List[Path]:
    """
    Will find all S1 patch directories in the provided `dir_path`.
    Only directories that strictly cohere to the naming convention will be returned.
    """
    return [
        p
        for p in dir_path.iterdir()
        if ben_constants.BEN_S1_RE.fullmatch(p.name) is not None
    ]


@functools.lru_cache()
def _conv_header_col_bz2_csv_resource_to_dict(
    resource, key_column: str, value_column: str
) -> Dict[str, str]:
    """
    Load a dictionary with the provided `key_column` and `value_column` after uncompressing
    the `bz2` compressed `resource` csv file.
    """
    # assert False, files(bigearthnet_common) / ""
    if not resources.is_resource(bigearthnet_common, resource):
        raise ValueError(
            f"{resource} resource is not available! This means that it was forgotten to be packaged."
        )

    with resources.path(bigearthnet_common, resource) as resource_path:
        with bz2.open(resource_path, mode="rt") as csv_file:
            reader = csv.DictReader(
                csv_file
            )  # field-names are encoded as first csv row
            if key_column not in reader.fieldnames:
                raise ValueError(
                    f"Key {key_column} is unkown! Resource provides: {reader.fieldnames}"
                )
            if value_column not in reader.fieldnames:
                raise ValueError(
                    f"Value {value_column} is unkown! Resource provides: {reader.fieldnames}"
                )
            return {row[key_column]: row[value_column] for row in reader}


# FUTURE: Could merge the underlying logic together
@functools.lru_cache()
def _conv_header_col_bz2_csv_resource_to_set(resource, key_column: str) -> Set[str]:
    """
    Load a dictionary with the provided `key_column` after uncompressing
    the `bz2` compressed `resource` csv file.
    """
    print(resource)
    if not resources.is_resource(bigearthnet_common, resource):
        # if not files(bigearthnet_common) / resource:
        raise ValueError(
            f"{resource} resource is not available! This means that it was forgotten to be packaged."
        )

    with resources.path(bigearthnet_common, resource) as resource_path:
        with bz2.open(resource_path, mode="rt") as csv_file:
            reader = csv.DictReader(
                csv_file
            )  # field-names are encoded as first csv row
            if key_column not in reader.fieldnames:
                raise ValueError(
                    f"Key {key_column} is unkown! Resource provides: {reader.fieldnames}"
                )
            return {row[key_column] for row in reader}


def get_all_s2_patch_names() -> Set[str]:
    resource = Resource.s1_s2_mapping
    return _conv_header_col_bz2_csv_resource_to_set(resource, "s2_name")


def get_all_s1_patch_names() -> Set[str]:
    resource = Resource.s1_s2_mapping
    return _conv_header_col_bz2_csv_resource_to_set(resource, "s1_name")


def _load_s1_s2_patch_name_mapping(from_s1_to_s2: bool = True) -> Dict[str, str]:
    """
    Load a dictionary which maps the S1 patch name to the S2 patch name (if `from_s1_to_s2`) or
    the S2 patch name to the S1 patch name.

    The compressed data could be regenerated with (requires the output of `bigearthnet_gdf_builder`):

    >>> import geopandas
    >>> raw_gdf = geopandas.read_parquet("raw_ben_s1_gdf.parquet")
    >>> raw_gdf = raw_gdf.rename({"name": "s1_name", "corresponding_s2_patch": "s2_name"}, axis=1)
    >>> raw_gdf.to_csv("s1_s2_mapping.csv.bz2", columns=["s1_name", "s2_name"], index=False)
    """
    resource = Resource.s1_s2_mapping
    key, value = ("s1_name", "s2_name") if from_s1_to_s2 else ("s2_name", "s1_name")
    return _conv_header_col_bz2_csv_resource_to_dict(resource, key, value)


def get_complete_s1_to_s2_patch_name_mapping() -> Dict[str, str]:
    """
    Load entire Sentinel-1 to Sentinel-2 BigEarthNet patch name mapping.

    Returns:
        Dict[str, str]: Sentinel-1 patch name keys with corresponding Sentinel-2 patch name as value
    """
    return _load_s1_s2_patch_name_mapping(from_s1_to_s2=True)


def get_complete_s2_to_s1_patch_name_mapping() -> Dict[str, str]:
    """
    Load entire Sentinel-2 to Sentinel-1 BigEarthNet patch name mapping.

    Returns:
        Dict[str, str]: Sentinel-2 patch name keys with corresponding Sentinel-1 patch name as value
    """
    return _load_s1_s2_patch_name_mapping(from_s1_to_s2=False)


def s1_to_s2_patch_name(s1_patch_name: str) -> str:
    """
    Convert BigEarthNet Sentinel-1 patch name to Sentinel-2 patch name.
    The function caches intermediate results.
    The function should be highly performant.

    Args:
        s1_patch_name (str): complete BigEarthNet Sentinel-1 patch name

    Returns:
        str: Corresponding Sentinel-2 patch name
    """
    return get_complete_s1_to_s2_patch_name_mapping()[s1_patch_name]


def s2_to_s1_patch_name(s2_patch_name: str) -> str:
    """
    Convert BigEarthNet Sentinel-2 patch name to Sentinel-1 patch name.
    The function caches intermediate results.
    The function should be highly performant.

    Args:
        s2_patch_name (str): complete BigEarthNet Sentinel-2 patch name

    Returns:
        str: Corresponding Sentinel-1 patch name
    """
    return get_complete_s2_to_s1_patch_name_mapping()[s2_patch_name]


def get_patches_to_country_mapping(use_s2_patch_names: bool = True) -> Dict[str, str]:
    """
    Return a dictionary that maps a patch name to a country.
    If `use_s2_patch_names` is set, use the BigEarthNet Sentinel-2 patch names.
    Otherwise, use the Sentinel-1 patch names.

    The compressed data could be regenerated with (requires the output of `bigearthnet_gdf_builder`):

    >>> import geopandas
    >>> extended_gdf = geopandas.read_parquet("extended_ben_s1_gdf.parquet")
    >>> extended_gdf = raw_gdf.rename({"name": "s1_name", "corresponding_s2_patch": "s2_name"}, axis=1)
    >>> extended_gdf.to_csv("country.csv.bz2", columns=["s1_name", "s2_name", "country"], index=False)
    """
    resource = Resource.country
    key = "s2_name" if use_s2_patch_names else "s1_name"
    return _conv_header_col_bz2_csv_resource_to_dict(resource, key, "country")


@validate_arguments
def is_s2_patch(patch_name: str) -> bool:
    """Quick regex check if name is a valid S2 patch name"""
    return ben_constants.BEN_S2_RE.fullmatch(patch_name)


@validate_arguments
def is_s1_patch(patch_name: str) -> bool:
    """Quick regex check if name is a valid S2 patch name"""
    return ben_constants.BEN_S1_RE.fullmatch(patch_name)


def get_country_from_patch_name(patch_name: str) -> str:
    """
    Fast function that returns the country to which the patch `patch_name` belongs to.

    This works for S1 and S2 patch names!
    """
    if is_s2_patch(patch_name):
        return get_patches_to_country_mapping(use_s2_patch_names=True)[patch_name]
    return get_patches_to_country_mapping(use_s2_patch_names=False)[patch_name]


def get_patches_to_season_mapping(use_s2_patch_names: bool = True) -> Dict[str, str]:
    """
    Return a dictionary that maps a patch name to the season of the acquisition date.
    If `use_s2_patch_names` is set, use the BigEarthNet Sentinel-2 patch names.
    Otherwise, use the Sentinel-1 patch names.

    The compressed data could be regenerated with (requires the output of `bigearthnet_gdf_builder`):

    >>> import geopandas
    >>> extended_gdf = geopandas.read_parquet("extended_ben_s1_gdf.parquet")
    >>> extended_gdf = raw_gdf.rename({"name": "s1_name", "corresponding_s2_patch": "s2_name"}, axis=1)
    >>> extended_gdf.to_csv("season.csv.bz2", columns=["s1_name", "s2_name", "season"], index=False)
    """
    resource = Resource.season
    key = "s2_name" if use_s2_patch_names else "s1_name"
    return _conv_header_col_bz2_csv_resource_to_dict(resource, key, "season")


def get_season_from_patch_name(patch_name: str) -> str:
    """
    Fast function that returns the season in which the patch `patch_name` was sensed.

    This works for S1 and S2 patch names!
    """
    if is_s2_patch(patch_name):
        return get_patches_to_season_mapping(use_s2_patch_names=True)[patch_name]
    return get_patches_to_season_mapping(use_s2_patch_names=False)[patch_name]


@validate_arguments
def _conv_single_col_csv_resource_to_set(
    resource: str,
) -> Set[str]:
    """
    Given a `resource` name of an encoded CSV file *without* a header
    line and only a single column, return the set of
    all values.
    """
    if not resources.is_resource(bigearthnet_common, resource):
        raise ValueError(f"{resource} is an unknown resource!")

    with resources.path(bigearthnet_common, resource) as resource_path:
        with bz2.open(resource_path, mode="rt") as csv_file:
            col_name = "Column"
            reader = csv.DictReader(csv_file, fieldnames=[col_name])
            return {row[col_name] for row in reader}


@functools.lru_cache()
def get_s2_patches_with_seasonal_snow() -> Set[str]:
    """List all patches with seasonal snow from **original** BigEarthNet-S2 dataset."""
    return _conv_single_col_csv_resource_to_set(Resource.patches_with_seasonal_snow)


@functools.lru_cache()
def get_s2_patches_with_cloud_and_shadow() -> Set[str]:
    """List all patches with cloud and shadow from **original** BigEarthNet-S2 dataset."""
    return _conv_single_col_csv_resource_to_set(Resource.patches_with_cloud_and_snow)


@functools.lru_cache()
def get_s1_patches_with_seasonal_snow() -> Set[str]:
    """List all patches with seasonal snow from **original** BigEarthNet-S1 dataset."""
    _s2_patches_with_snow = get_s2_patches_with_seasonal_snow()
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in _s2_patches_with_snow}


@functools.lru_cache()
def get_s1_patches_with_cloud_and_shadow() -> Set[str]:
    """List all patches with cloud and shadow from **original** BigEarthNet-S1 dataset."""
    _s2_patches_with_clouds = get_s2_patches_with_cloud_and_shadow()
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in _s2_patches_with_clouds}


@validate_arguments
def is_snowy_patch(patch_name: str):
    """
    Fast function that checks whether `patch_name` is a patch
    that contains a lot of seasonal snow.

    This works for S1 and S2 patch names!
    """
    return (
        patch_name in get_s2_patches_with_seasonal_snow()
        or patch_name in get_s1_patches_with_seasonal_snow()
    )


@validate_arguments
def is_cloudy_shadowy_patch(patch_name: str):
    """
    Fast function that checks whether `patch_name` is a patch
    that contains a lot of shadow or is obstructed by clouds.

    This works for S1 and S2 patch names!
    """
    return (
        patch_name in get_s2_patches_with_cloud_and_shadow()
        or patch_name in get_s1_patches_with_cloud_and_shadow()
    )


@functools.lru_cache()
def get_s2_patches_with_no_19_class_target() -> Set[str]:
    """
    List all patches from the BigEarthNet-S2 dataset that
    have _no_ defined classes with the 19-class nomenclature.

    Note: This set still includes patches with snow, clouds, or shadows.

    To re-build the file, it is necessary to use the output of `bigearthnet_gdf_builder`:

    >>> raw_gdf = geopandas.read_parquet("raw_ben_s2_gdf.parquet")
    >>> raw_gdf["new_labels"] = raw_gdf["labels"].apply(old2new_labels)
    >>> no_19_label_targets = raw_gdf[raw_gdf["new_labels"].isna()]
    >>> no_19_label_targets.to_csv("patches_with_no_19_class_targets.csv.bz2", columns=["name"], index=False, header=False)
    """
    return _conv_single_col_csv_resource_to_set(
        Resource.patches_with_no_19_class_targets
    )


@functools.lru_cache()
def get_s1_patches_with_no_19_class_target() -> Set[str]:
    """
    List all patches from the BigEarthNet-S1 dataset that
    have _no_ defined classes with the 19-class nomenclature.

    Note: This set still includes patches with snow, clouds, or shadows.

    The patch names are converted from `get_s2patches_with_no_19_class_targets`
    for compactness.
    """
    s2_patches_no_19_classes = _conv_single_col_csv_resource_to_set(
        Resource.patches_with_no_19_class_targets
    )
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in s2_patches_no_19_classes}


@validate_arguments
def has_19_class_target(patch_name: str) -> bool:
    """
    Fast function that checks whether `patch_name` is a patch
    that has at least a single 19-class target label.

    This works for S1 and S2 patch names!
    """
    return not (
        patch_name in get_s1_patches_with_no_19_class_target()
        or patch_name in get_s2_patches_with_no_19_class_target()
    )


# FUTURE: Remove this bz2 file and repackage it inside of the
# metadata collection file
# "https://git.tu-berlin.de/rsim/BigEarthNet-S2_19-classes_models/-/raw/master/splits/train.csv",
@functools.lru_cache()
def get_s2_patches_from_original_train_split() -> Set[str]:
    """
    List all Sentinel-2 train patches from the original train/validation/test split.
    """
    return _conv_single_col_csv_resource_to_set(Resource.train_csv)


@functools.lru_cache()
def get_s1_patches_from_original_train_split() -> Set[str]:
    """
    List all Sentinel-1 train patches from the original train/validation/test split.
    """
    s2_train_patches = get_s2_patches_from_original_train_split()
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in s2_train_patches}


@functools.lru_cache()
def get_s2_patches_from_original_validation_split() -> Set[str]:
    """
    List all Sentinel-2 validation patches from the original train/validation/test split.
    """
    return _conv_single_col_csv_resource_to_set(Resource.val_csv)


@functools.lru_cache()
def get_s1_patches_from_original_validation_split() -> Set[str]:
    """
    List all Sentinel-1 validation patches from the original train/validation/test split.
    """
    s2_validation_patches = get_s2_patches_from_original_validation_split()
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in s2_validation_patches}


@functools.lru_cache()
def get_s2_patches_from_original_test_split() -> Set[str]:
    """
    List all Sentinel-2 test patches from the original train/validation/test split.
    """
    return _conv_single_col_csv_resource_to_set(Resource.test_csv)


@functools.lru_cache()
def get_s1_patches_from_original_test_split() -> Set[str]:
    """
    List all Sentinel-1 test patches from the original train/validation/test split.
    """
    s2_test_patches = get_s2_patches_from_original_test_split()
    return {s2_to_s1_patch_name(s2_patch) for s2_patch in s2_test_patches}


@validate_arguments
def get_original_split_from_patch_name(patch: str) -> Optional[str]:
    """
    Returns "train"/"validation"/"test" or `None`.
    The value is retrieved from the original BigEarthNet-S1/S2
    train/validation/test split. If the input is not present
    in any split, it will return `None` and raise a UserWarning.
    This happens for patches that are either in the
    cloud/shadow or seasonal snow set or if there exists no 19-label target.

    Note: This works for Sentinel-2 and Sentinel-1 patch names!
    """
    s1_train = get_s1_patches_from_original_train_split()
    s2_train = get_s2_patches_from_original_train_split()
    s1_validation = get_s1_patches_from_original_validation_split()
    s2_validation = get_s2_patches_from_original_validation_split()
    s1_test = get_s1_patches_from_original_test_split()
    s2_test = get_s2_patches_from_original_test_split()

    if patch in s1_train or patch in s2_train:
        return ben_constants.Split.train
    elif patch in s1_validation or patch in s2_validation:
        return ben_constants.Split.validation
    elif patch in s1_test or patch in s2_test:
        return ben_constants.Split.test
    warnings.warn(
        "Provided an input patch name which was not part of the original split.",
        UserWarning,
    )
    return None


@validate_arguments
def _old2new_label(old_label: str) -> Optional[str]:
    """
    Converts old-style BigEearthNet label to the
    new labels.

    > Note: Some labels were removed! This function
    will return `None` if the label was removed and
    raise a `KeyError` if the input label is unknown.
    """
    return ben_constants.OLD2NEW_LABELS_DICT[old_label]


def old2new_labels(old_labels: Iterable[str]) -> Optional[List[str]]:
    """
    Converts a list of old-style BigEarthNet labels
    to a list of labels.

    If there are no corresponding new labels (which can happen with original BEN patches!)
    then the function will return `None` and raise a user warning.

    If an illegal/unknown input label is provided, a `KeyError` is raised.
    """
    new_labels = [
        _old2new_label(l) for l in old_labels if _old2new_label(l) is not None
    ]
    if len(old_labels) > 0 and len(new_labels) == 0:
        warnings.warn(
            "Provided a list of old labels that only contains `removed` labels!",
            UserWarning,
        )
        new_labels = None
    return new_labels


@validate_arguments
def ben_19_labels_to_multi_hot(
    labels: Iterable[str], lex_sorted: bool = True
) -> List[float]:
    """
    Convenience function that converts an input iterable of labels into
    a multi-hot encoded vector.
    If `lex_sorted` is True (default) the classes are lexigraphically ordered, as they are
    in `constants.NEW_LABELS`.
    If `lex_sorted` is False, the original order from the BigEarthNet paper is used, as
    they are given in `constants.NEW_LABELS_ORIGINAL_ORDER`.

    If an unknown label is given, a `KeyError` is raised.

    Be aware that this approach assumes that **all** labels are actually used in the dataset!
    This is not necessarily the case if you are using a subset!
    For example, the "Agro-forestry areas" class is only present in Portugal and in no other country!
    """
    ordered_lbls = (
        ben_constants.NEW_LABELS
        if lex_sorted
        else ben_constants.NEW_LABELS_ORIGINAL_ORDER
    )
    lbls_to_idx = fc.L(ordered_lbls).val2idx()
    idxs = [lbls_to_idx[label] for label in labels]
    multi_hot = fc.L([0] * len(ben_constants.NEW_LABELS))
    multi_hot[idxs] = 1.0
    return list(multi_hot)


@validate_arguments
def ben_43_labels_to_multi_hot(
    labels: Iterable[str], lex_sorted: bool = True
) -> List[float]:
    """
    Convenience function that converts an input iterable of labels into
    a multi-hot encoded vector.
    If `lex_sorted` is True (default) the classes are lexigraphically ordered, as they are
    in `constants.OLD_LABELS`.
    If `lex_sorted` is False, the original order from the BigEarthNet paper is used, as
    they are given in `constants.OLD_LABELS_ORIGINAL_ORDER`.

    If an unknown label is given, a `KeyError` is raised.

    Be aware that this approach assumes that **all** labels are actually used in the dataset!
    This is not necessarily the case if you are using a subset!
    For example, the "Agro-forestry areas" class is only present in Portugal and in no other country!
    """
    ordered_lbls = (
        ben_constants.OLD_LABELS
        if lex_sorted
        else ben_constants.OLD_LABELS_ORIGINAL_ORDER
    )
    lbls_to_idx = fc.L(ordered_lbls).val2idx()
    idxs = [lbls_to_idx[label] for label in labels]
    multi_hot = fc.L([0] * len(ben_constants.OLD_LABELS))
    multi_hot[idxs] = 1.0
    return list(multi_hot)


def _are_s1_files_complete(patch_path: DirectoryPath) -> bool:
    """
    Check if all S1-patch files exists (bands and json files) and are not empty.
    """
    file_suffixes = ["VV", "VH", "_labels_metadata.json"]
    for suffix in file_suffixes:
        file = patch_path / f"{patch_path.name}{suffix}"
        if not file.exists() or file.stat().st_size == 0:
            return False
    return True


def _are_s2_files_complete(patch_path: DirectoryPath) -> bool:
    """
    Check if all S2-patch files exists (bands and json files) and are not empty.
    """
    file_suffixes = [f"_B{i:02}.tif" for i in range(1, 13) if i != 10] + [
        "_B8A",
        "_labels_metadata.json",
    ]
    for suffix in file_suffixes:
        file = patch_path / f"{patch_path.name}{suffix}"
        if not file.exists() or file.stat().st_size == 0:
            return False
    return True


def _print_missing_dirs(missing_dirs, show_num: int = 10) -> None:
    rich.print("There are some missing directories!")
    rich.print(
        "The following directories are missing compared to the complete BEN archive."
    )
    show_num = min(show_num, len(missing_dirs))
    rich.print(
        f"Showing only the first {show_num} invalid directories of {len(missing_dirs)}"
    )
    rich.print([d for _, d in zip(range(show_num), missing_dirs)])


def _print_dirs_with_missing_files(dirs_with_missing_files, show_num: int = 10) -> None:
    rich.print("There are some invalid directories!")
    rich.print("The following directories are missing files.")
    show_num = min(show_num, len(dirs_with_missing_files))
    rich.print(
        f"Showing only the first {show_num} invalid directories of {len(dirs_with_missing_files)}"
    )
    rich.print([d for _, d in zip(range(show_num), dirs_with_missing_files)])


@validate_arguments
def _validate_ben_root_directory(
    dir_path: DirectoryPath, is_sentinel2: bool
) -> Set[str]:
    files = {f for f in dir_path.glob("*")}
    patch_names = get_all_s2_patch_names() if is_sentinel2 else get_all_s1_patch_names()
    expected_directories = {dir_path / patch for patch in patch_names}
    missing_directories = expected_directories - files
    ben_dirs = expected_directories & files
    completeness_checker = (
        _are_s2_files_complete if is_sentinel2 else _are_s1_files_complete
    )
    directories_with_missing_files = {
        f for f in ben_dirs if not completeness_checker(f)
    }
    if missing_directories == set() and directories_with_missing_files == set():
        rich.print("Nothing seems to be missing.")
        rich.print(f"The Sentinel directory {dir_path} looks complete.")
        return set()
    if missing_directories != set():
        _print_missing_dirs(missing_directories)
    if directories_with_missing_files != set():
        _print_dirs_with_missing_files(directories_with_missing_files)
    return missing_directories | directories_with_missing_files


def validate_ben_s2_root_directory(dir_path: Path) -> Set[str]:
    """
    Quickly check if all expected files from the BigEarthNet-S2 archive are present.

    This funtion will _not_ check if the files are correct or if they were modified!
    The function will perform a simple existence check and verify that each file is not empty.
    Other files will be ignored.
    """
    return _validate_ben_root_directory(dir_path, is_sentinel2=True)


def validate_ben_s1_root_directory(dir_path: Path) -> Set[str]:
    """
    Quickly check if all expected files from the BigEarthNet-S1 archive are present.

    This funtion will _not_ check if the files are correct or if they were modified!
    The function will perform a simple existence check and verify that each file is not empty.
    Other files will be ignored.
    """
    return _validate_ben_root_directory(dir_path, is_sentinel2=False)


def validate_ben_s2_root_directory_cli():
    app = typer.Typer()
    app.command()(validate_ben_s2_root_directory)
    app()


def validate_ben_s1_root_directory_cli():
    app = typer.Typer()
    app.command()(validate_ben_s1_root_directory)
    app()


@validate_arguments
def describe_patch(patch_names: List[str]) -> Table:
    """
    Given a list of patch names return a table that summarizes
    the metadata.
    """
    columns = [
        "Sentinel-1 Name",
        "Sentinel-2 Name",
        "Original Split",
        "Country",
        "Season",
        "Snowy",
        "Cloudy / Shadowy",
        "Valid 19-label",
    ]
    y = "✅"
    n = "❌"
    t = Table(
        title=f"Metadata Summary",
        caption=f"Patch(es): {patch_names}",
        expand=True,
        leading=1,
    )
    for c in columns:
        t.add_column(header=c, overflow="fold", justify="center")

    for patch_name in patch_names:
        if is_s1_patch(patch_name):
            s1_name = patch_name
            s2_name = s1_to_s2_patch_name(s1_name)
        elif is_s2_patch(patch_name):
            s2_name = patch_name
            s1_name = s2_to_s1_patch_name(s2_name)
        else:
            raise ValueError(f"Input: {patch_name} is not a valid S1/S2 patch name!")

        split = get_original_split_from_patch_name(s2_name)
        country = get_country_from_patch_name(s2_name)
        season = get_season_from_patch_name(s2_name)
        snowy = y if is_snowy_patch(s2_name) else n
        cloudy = y if is_cloudy_shadowy_patch(s2_name) else n
        valid_19 = y if has_19_class_target(s2_name) else n
        t.add_row(s1_name, s2_name, split, country, season, snowy, cloudy, valid_19)
    rich.print(t)
    return t


def describe_patch_cli():
    app = typer.Typer()
    app.command()(describe_patch)
    app()



import csv
from pathlib import Path
from typing import List, Sequence, Set

import fastcore.all as fc
import natsort
import rich_click as click
import rich_click.typer as typer
from pydantic import validate_arguments

click.rich_click.USE_MARKDOWN = True

import bigearthnet_common.base as ben_base
import bigearthnet_common.constants as ben_constants


def get_all_s2_patches() -> Set[str]:
    """
    Returns a set of all S2 patch names.
    """
    return set(ben_base.get_complete_s2_to_s1_patch_name_mapping().keys())


def get_all_s1_patches() -> Set[str]:
    """
    Returns a set of all S1 patch names.
    """
    return set(ben_base.get_complete_s1_to_s2_patch_name_mapping().keys())


def get_recommended_s2_patches() -> Set[str]:
    s2_patches = get_all_s2_patches()
    no_19_class_targets = ben_base.get_s2_patches_with_no_19_class_target()
    recommended_s2_patches = {
        p
        for p in s2_patches
        if not ben_base.is_snowy_patch(p)
        and not ben_base.is_cloudy_shadowy_patch(p)
        and p not in no_19_class_targets
    }
    return recommended_s2_patches


def get_recommended_s1_patches() -> Set[str]:
    s1_patches = get_all_s1_patches()
    no_19_class_targets = ben_base.get_s1_patches_with_no_19_class_target()
    recommended_s1_patches = {
        p
        for p in s1_patches
        if not ben_base.is_snowy_patch(p)
        and not ben_base.is_cloudy_shadowy_patch(p)
        and p not in no_19_class_targets
    }
    return recommended_s1_patches


@validate_arguments
def filter_s2_patches_by_country(
    patches: Sequence, country: ben_constants.Country
) -> Set[str]:
    if country not in ben_constants.COUNTRIES:
        raise ValueError(
            f"{country} is not one of the BEN countries: {ben_constants.COUNTRIES}!"
        )

    patch_country_mapping = ben_base.get_patches_to_country_mapping(
        use_s2_patch_names=True
    )
    return {p for p in patches if patch_country_mapping[p] == country}


@validate_arguments
def filter_s1_patches_by_country(
    patches: Sequence, country: ben_constants.Country
) -> Set[str]:
    if country not in ben_constants.COUNTRIES:
        raise ValueError(
            f"{country} is not one of the BEN countries: {ben_constants.COUNTRIES}!"
        )

    patch_country_mapping = ben_base.get_patches_to_country_mapping(
        use_s2_patch_names=False
    )
    return {p for p in patches if patch_country_mapping[p] == country}


@validate_arguments
def filter_patches_by_country(
    sentinel_source: ben_constants.SentinelSource,
    patches: Sequence,
    country: ben_constants.Country,
):
    """
    Given Sentinel-1/2 named-patches, return only those patches that belong to a given
    country.
    """
    return (
        filter_s1_patches_by_country(patches, country)
        if sentinel_source == ben_constants.SentinelSource.S1
        else filter_s2_patches_by_country(patches, country)
    )


@validate_arguments
def filter_s2_patches_by_season(
    patches: Sequence, season: ben_constants.Season
) -> Set[str]:
    patch_season_mapping = ben_base.get_patches_to_season_mapping(
        use_s2_patch_names=True
    )
    return {p for p in patches if patch_season_mapping[p] == season}


@validate_arguments
def filter_s1_patches_by_season(
    patches: Sequence, season: ben_constants.Season
) -> Set[str]:
    patch_season_mapping = ben_base.get_patches_to_season_mapping(
        use_s2_patch_names=False
    )
    return {p for p in patches if patch_season_mapping[p] == season}


@validate_arguments
def filter_patches_by_season(
    sentinel_source: ben_constants.SentinelSource,
    patches: Sequence,
    season: ben_constants.Season,
):
    """
    Given Sentinel-1/2 named-patches, return only those patches that belong to a given
    season.
    """
    return (
        filter_s1_patches_by_season(patches, season)
        if sentinel_source == ben_constants.SentinelSource.S1
        else filter_s2_patches_by_season(patches, season)
    )


@validate_arguments
def filter_s1_patches_by_split(patches: Sequence, split: ben_constants.Split):
    get_split_func = {
        split.train: ben_base.get_s1_patches_from_original_train_split,
        split.validation: ben_base.get_s1_patches_from_original_validation_split,
        split.test: ben_base.get_s1_patches_from_original_test_split,
    }
    split_patches = get_split_func[split]()
    return patches & split_patches


@validate_arguments
def filter_s2_patches_by_split(patches: Sequence, split: ben_constants.Split):
    get_split_func = {
        split.train: ben_base.get_s2_patches_from_original_train_split,
        split.validation: ben_base.get_s2_patches_from_original_validation_split,
        split.test: ben_base.get_s2_patches_from_original_test_split,
    }
    split_patches = get_split_func[split]()
    return set(patches) & split_patches


@validate_arguments
def filter_patches_by_split(
    sentinel_source: ben_constants.SentinelSource,
    patches: Sequence,
    split: ben_constants.Split,
):
    """
    Given Sentinel-1/2 named-patches, return only those patches that belong to a given
    split.
    """
    return (
        filter_s1_patches_by_split(patches, split)
        if sentinel_source == ben_constants.SentinelSource.S1
        else filter_s2_patches_by_split(patches, split)
    )


@validate_arguments
def build_set(
    sentinel_source: ben_constants.SentinelSource,
    seasons: List[ben_constants.Season] = [s.value for s in ben_constants.Season],
    countries: List[ben_constants.Country] = [c.value for c in ben_constants.Country],
    remove_unrecommended_dl_patches: bool = True,
) -> Set[str]:
    """
    Create a subset of the Sentinel-1/Sentinel-2 patches.
    The returned list will be naturally sorted to produce
    deterministic results.
    """
    use_s1 = sentinel_source == ben_constants.SentinelSource.S1
    # FUTURE: could be split up into higher level functions
    if remove_unrecommended_dl_patches:
        patches = (
            get_recommended_s1_patches() if use_s1 else get_recommended_s2_patches()
        )
    else:
        patches = get_all_s1_patches() if use_s1 else get_all_s2_patches()

    if len(seasons) > 0:
        patches = {
            patch
            for season in seasons
            for patch in filter_patches_by_season(sentinel_source, patches, season)
        }
    if len(countries) > 0:
        patches = {
            patch
            for country in countries
            for patch in filter_patches_by_country(sentinel_source, patches, country)
        }
    return patches


@fc.delegates(build_set)
@validate_arguments
def build_csv_sets(
    file_path: Path,
    sentinel_source: ben_constants.SentinelSource,
    write_separate_splits: bool = True,
    **kwargs,
):
    """
    Build CSV files that contain Sentinel-1/2 patches names with the given restrictions.
    By default, the patches will be grouped by the orignal train/validation/test split.
    This is generally a good starting point, but not ideal for all use-cases.
    If `write_separate_splits` is `False`, the output will contain all patches.
    The generated file does **not** contain a header row, as this style was used in all the other
    publicly available BigEarthNet CSV files.

    The patch names will be naturally sorted to ensure deterministic outputs.
    """

    def _write_csv(fp, patches):
        sorted_patches = natsort.natsorted(patches)
        with open(fp.with_suffix(".csv"), "w") as csv_file:
            writer = csv.writer(csv_file)
            for patch in sorted_patches:
                writer.writerow([patch])

    patches = build_set(sentinel_source, **kwargs)
    if write_separate_splits:
        train_patches = filter_patches_by_split(
            sentinel_source, patches, ben_constants.Split.train
        )
        validation_patches = filter_patches_by_split(
            sentinel_source, patches, ben_constants.Split.validation
        )
        test_patches = filter_patches_by_split(
            sentinel_source, patches, ben_constants.Split.test
        )
        if min(len(train_patches), len(validation_patches), len(test_patches)) == 0:
            raise RuntimeError("One of the train/val/test splits is empty!")
        if (train_patches & validation_patches & test_patches) != set():
            raise RuntimeError(
                "There is an overlap between the train/validation/test patches!"
            )
        _write_csv(file_path.with_name(f"{file_path.name}_train.csv"), train_patches)
        _write_csv(
            file_path.with_name(f"{file_path.name}_validation.csv"), validation_patches
        )
        _write_csv(file_path.with_name(f"{file_path.name}_test.csv"), test_patches)
    else:
        _write_csv(file_path, patches)


def build_csv_sets_cli():
    app = typer.Typer(name="ben_build_csv_sets")
    app.command()(build_csv_sets)
    app()


if __name__ == "__main__" and not fc.IN_IPYTHON:
    build_csv_sets_cli()



import re
from enum import Enum

import natsort
import rich


def _generate_old2new_labels_dict():
    """
    Code used to generate Old2New_labels dictionary.
    Only run once and saved the output as a constant.
    """
    import requests

    src_url = "https://git.tu-berlin.de/rsim/BigEarthNet-S2_19-classes_models/-/raw/master/label_indices.json"
    r = requests.get(src_url)
    src = r.json()
    original_labels = list(src["original_labels"].keys())
    assert len(original_labels) == 43
    new_labels = list(src["BigEarthNet-19_labels"].keys())
    assert len(new_labels) == 19
    label_conv = src["label_conversion"]
    assert len(label_conv) == len(new_labels)

    # d = {}
    # for new_label_idx, old_label_idxs in enumerate(label_conv):
    #     for old_label_idx in old_label_idxs:
    #         d[original_labels[old_label_idx]] = new_labels[new_label_idx]

    # print(d)

    old2new_label_mapping = {
        original_labels[old_idx]: new_labels[new_label_idx]
        for new_label_idx, row in enumerate(src["label_conversion"])
        for old_idx in row
    }

    deleted_labels = set(original_labels) - set(old2new_label_mapping.keys())

    for l in deleted_labels:
        old2new_label_mapping[l] = None

    assert len(old2new_label_mapping.keys()) == len(original_labels)

    return old2new_label_mapping


# Same order as in the original BigEarthNet code:
# https://git.tu-berlin.de/rsim/BigEarthNet-S2_19-classes_models/-/raw/master/label_indices.json
OLD2NEW_LABELS_DICT = {
    "Continuous urban fabric": "Urban fabric",
    "Discontinuous urban fabric": "Urban fabric",
    "Industrial or commercial units": "Industrial or commercial units",
    "Road and rail networks and associated land": None,
    "Port areas": None,
    "Airports": None,
    "Mineral extraction sites": None,
    "Dump sites": None,
    "Construction sites": None,
    "Green urban areas": None,
    "Sport and leisure facilities": None,
    "Non-irrigated arable land": "Arable land",
    "Permanently irrigated land": "Arable land",
    "Rice fields": "Arable land",
    "Vineyards": "Permanent crops",
    "Fruit trees and berry plantations": "Permanent crops",
    "Olive groves": "Permanent crops",
    "Pastures": "Pastures",
    "Annual crops associated with permanent crops": "Permanent crops",
    "Complex cultivation patterns": "Complex cultivation patterns",
    "Land principally occupied by agriculture, with significant areas of natural vegetation": "Land principally occupied by agriculture, with significant areas of natural vegetation",
    "Agro-forestry areas": "Agro-forestry areas",
    "Broad-leaved forest": "Broad-leaved forest",
    "Coniferous forest": "Coniferous forest",
    "Mixed forest": "Mixed forest",
    "Natural grassland": "Natural grassland and sparsely vegetated areas",
    "Moors and heathland": "Moors, heathland and sclerophyllous vegetation",
    "Sclerophyllous vegetation": "Moors, heathland and sclerophyllous vegetation",
    "Transitional woodland/shrub": "Transitional woodland, shrub",
    "Beaches, dunes, sands": "Beaches, dunes, sands",
    "Bare rock": None,
    "Sparsely vegetated areas": "Natural grassland and sparsely vegetated areas",
    "Burnt areas": None,
    "Inland marshes": "Inland wetlands",
    "Peatbogs": "Inland wetlands",
    "Salt marshes": "Coastal wetlands",
    "Salines": "Coastal wetlands",
    "Intertidal flats": None,
    "Water courses": "Inland waters",
    "Water bodies": "Inland waters",
    "Coastal lagoons": "Marine waters",
    "Estuaries": "Marine waters",
    "Sea and ocean": "Marine waters",
}


# Use sorted order as default => Same encoding as used in torchgeo
# https://torchgeo.readthedocs.io/en/latest/api/datasets.html#bigearthnet
OLD_LABELS_ORIGINAL_ORDER = tuple(k for k in OLD2NEW_LABELS_DICT.keys())
OLD_LABELS = sorted(OLD_LABELS_ORIGINAL_ORDER)

# The following would break the original order
# NEW_LABELS_ORIGINAL_ORDER = tuple(
#     {v for v in OLD2NEW_LABELS_DICT.values() if v is not None}
# )
NEW_LABELS_ORIGINAL_ORDER = (
    "Urban fabric",
    "Industrial or commercial units",
    "Arable land",
    "Permanent crops",
    "Pastures",
    "Complex cultivation patterns",
    "Land principally occupied by agriculture, with significant areas of natural vegetation",
    "Agro-forestry areas",
    "Broad-leaved forest",
    "Coniferous forest",
    "Mixed forest",
    "Natural grassland and sparsely vegetated areas",
    "Moors, heathland and sclerophyllous vegetation",
    "Transitional woodland, shrub",
    "Beaches, dunes, sands",
    "Inland wetlands",
    "Coastal wetlands",
    "Inland waters",
    "Marine waters",
)
NEW_LABELS = sorted(NEW_LABELS_ORIGINAL_ORDER)


# Manually copied the LVL3 labels to be able to test for typos from both sources
# Strings copied from
# https://land.copernicus.eu/user-corner/technical-library/corine-land-cover-nomenclature-guidelines/html/index.html
CLC_LV3_TO_LV2 = {
    "Continuous urban fabric": "Urban fabric",
    "Discontinuous urban fabric": "Urban fabric",
    "Industrial or commercial units": "Industrial, comercial and transport units",
    "Road and rail networks and associated land": "Industrial, comercial and transport units",
    "Port areas": "Industrial, comercial and transport units",
    "Airports": "Industrial, comercial and transport units",
    "Mineral extraction sites": "Mine, dump and construction sites",
    "Dump sites": "Mine, dump and construction sites",
    "Construction sites": "Mine, dump and construction sites",
    "Green urban areas": "Artificial, non-agricultural vegetated areas",
    "Sport and leisure facilities": "Artificial, non-agricultural vegetated areas",
    "Non-irrigated arable land": "Arable land",
    "Permanently irrigated land": "Arable land",
    "Rice fields": "Arable land",
    "Vineyards": "Permanent crops",
    "Fruit trees and berry plantations": "Permanent crops",
    "Olive groves": "Permanent crops",
    "Pastures": "Pastures",
    "Annual crops associated with permanent crops": "Heterogeneous agricultural areas",
    "Complex cultivation patterns": "Heterogeneous agricultural areas",
    "Land principally occupied by agriculture, with significant areas of natural vegetation": "Heterogeneous agricultural areas",
    "Agro-forestry areas": "Heterogeneous agricultural areas",
    "Broad-leaved forest": "Forest",
    "Coniferous forest": "Forest",
    "Mixed forest": "Forest",
    "Natural grassland": "Shrub and/or herbaceous vegetation associations",
    "Moors and heathland": "Shrub and/or herbaceous vegetation associations",
    "Sclerophyllous vegetation": "Shrub and/or herbaceous vegetation associations",
    "Transitional woodland/shrub": "Shrub and/or herbaceous vegetation associations",
    "Beaches, dunes, sands": "Open spaces with little or no vegetation",
    "Bare rock": "Open spaces with little or no vegetation",
    "Sparsely vegetated areas": "Open spaces with little or no vegetation",
    "Burnt areas": "Open spaces with little or no vegetation",
    "Glaciers and perpetual snow": "Open spaces with little or no vegetation",
    "Inland marshes": "Inland wetlands",
    "Peatbogs": "Inland wetlands",
    "Salt marshes": "Coastal wetlands",
    "Salines": "Coastal wetlands",
    "Intertidal flats": "Coastal wetlands",
    "Water courses": "Inland waters",
    "Water bodies": "Inland waters",
    "Coastal lagoons": "Marine waters",
    "Estuaries": "Marine waters",
    "Sea and ocean": "Marine waters",
}

CLC_LV2_TO_LV1 = {
    "Urban fabric": "Artificial Surfaces",
    "Industrial, comercial and transport units": "Artificial Surfaces",
    "Mine, dump and construction sites": "Artificial Surfaces",
    "Artificial, non-agricultural vegetated areas": "Artificial Surfaces",
    "Arable land": "Agricultural areas",
    "Permanent crops": "Agricultural areas",
    "Pastures": "Agricultural areas",
    "Heterogeneous agricultural areas": "Agricultural areas",
    "Forest": "Forest and seminatural areas",
    "Shrub and/or herbaceous vegetation associations": "Forest and seminatural areas",
    "Open spaces with little or no vegetation": "Forest and seminatural areas",
    "Inland wetlands": "Wetlands",
    "Coastal wetlands": "Wetlands",
    "Inland waters": "Water bodies",
    "Marine waters": "Water bodies",
}

CLC_LV3_TO_LV1 = {lv3: CLC_LV2_TO_LV1[lv2] for lv3, lv2 in CLC_LV3_TO_LV2.items()}
CLC_LV1_LABELS = (
    "Artificial Surfaces",
    "Agricultural areas",
    "Forest and seminatural areas",
    "Wetlands",
    "Water bodies",
)

CLC_LV3_LABELS = tuple(CLC_LV3_TO_LV2.keys())
CLC_LV2_LABELS = tuple(CLC_LV2_TO_LV1.keys())


# manually added to double-check values
CLC_LV3_COUNT = 44
CLC_LV2_COUNT = 15
CLC_LV1_COUNT = 5


CLC_LV1_TO_CLC_CODE = {
    "Artificial Surfaces": 1,
    "Agricultural areas": 2,
    "Forest and seminatural areas": 3,
    "Wetlands": 4,
    "Water bodies": 5,
}
CLC_CODE_TO_CLC_LV1 = {v: k for k, v in CLC_LV1_TO_CLC_CODE.items()}

CLC_LV2_TO_CLC_CODE = {
    "Urban fabric": 11,
    "Industrial, comercial and transport units": 12,
    "Mine, dump and construction sites": 13,
    "Artificial, non-agricultural vegetated areas": 14,
    "Arable land": 21,
    "Permanent crops": 22,
    "Pastures": 23,
    "Heterogeneous agricultural areas": 24,
    "Forest": 31,
    "Shrub and/or herbaceous vegetation associations": 32,
    "Open spaces with little or no vegetation": 33,
    "Inland wetlands": 41,
    "Coastal wetlands": 42,
    "Inland waters": 51,
    "Marine waters": 52,
}
CLC_CODE_TO_CLC_LV2 = {v: k for k, v in CLC_LV2_TO_CLC_CODE.items()}

# Requires knowledge that dict keep ordering in supported Python version
# as well as the correct order in the mappings
CLC_LV3_TO_CLC_CODE = {
    lv3_lbl: f"{CLC_LV2_TO_CLC_CODE[lv2_lbl]}{i}"
    for lv2_lbl in CLC_LV3_TO_LV2.values()
    for i, lv3_lbl in enumerate(
        [k for k, v in CLC_LV3_TO_LV2.items() if v == lv2_lbl], 1
    )
}

CLC_CODE_TO_CLC_LV3 = {v: k for k, v in CLC_LV3_TO_CLC_CODE.items()}


# Inspired by
# https://github.com/albumentations-team/albumentations/blob/master/albumentations/augmentations/functional.py#L1259
MAX_VALUES_BY_DTYPE_STR = {
    "uint8": 255,
    "uint16": 65535,
    "uint32": 4294967295,
    "float32": 1.0,
    "float64": 1.0,
}


URL = "http://bigearth.net/downloads/BigEarthNet-v1.0.tar.gz"

# Stats from https://git.tu-berlin.de/rsim/bigearthnet-models-tf/-/blob/master/BigEarthNet.py
BAND_STATS_S2 = {
    "mean": {
        "B01": 340.76769064,
        "B02": 429.9430203,
        "B03": 614.21682446,
        "B04": 590.23569706,
        "B05": 950.68368468,
        "B06": 1792.46290469,
        "B07": 2075.46795189,
        "B08": 2218.94553375,
        "B8A": 2266.46036911,
        "B09": 2246.0605464,
        "B11": 1594.42694882,
        "B12": 1009.32729131,
    },
    "std": {
        "B01": 554.81258967,
        "B02": 572.41639287,
        "B03": 582.87945694,
        "B04": 675.88746967,
        "B05": 729.89827633,
        "B06": 1096.01480586,
        "B07": 1273.45393088,
        "B08": 1365.45589904,
        "B8A": 1356.13789355,
        "B09": 1302.3292881,
        "B11": 1079.19066363,
        "B12": 818.86747235,
    },
}
BAND_STATS_S1 = {
    "mean": {
        "VV": -12.619993741972035,
        "VH": -19.29044597721542,
        "VV/VH": 0.6525036195871579,
    },
    "std": {
        "VV": 5.115911777546365,
        "VH": 5.464428464912864,
        "VV/VH": 30.75264076801808,
    },
}

# I am not sure if this should be used!
# FUTURE: Double check original values!
# Custom transformation to float32 values.
# Assumes that the original means and std values are encoded in uint16
# BAND_STATS_FLOAT32 = {
#     k: {
#         band: band_val / MAX_VALUES_BY_DTYPE_STR["uint16"]
#         for band, band_val in v.items()
#     }
#     for k, v in BAND_STATS.items()
# }

BEN_CHANNELS = (
    "B01",
    "B02",
    "B03",
    "B04",
    "B05",
    "B06",
    "B07",
    "B08",
    "B8A",
    "B09",
    "B11",
    "B12",
)

BEN_10m_CHANNELS = (
    "B02",
    "B03",
    "B04",
    "B08",
)

BEN_20m_CHANNELS = (
    "B05",
    "B06",
    "B07",
    "B8A",
    "B11",
    "B12",
)

BEN_10m_20m_CHANNELS = natsort.natsorted(BEN_10m_CHANNELS + BEN_20m_CHANNELS)

BEN_60m_CHANNELS = (
    "B01",
    "B09",
)

# Correctly ordered
BEN_RGB_CHANNELS = (
    "B04",
    "B03",
    "B02",
)

BEN_PATCH_SIZE_M = 1200


BEN_S1_V1_0_JSON_KEYS = {
    "acquisition_time",
    "coordinates",
    "labels",
    "projection",
    "scene_source",
    "corresponding_s2_patch",
}

# note that there is a small difference between S1/S2
# acquisition_time vs acquisition_date
BEN_S2_V1_0_JSON_KEYS = {
    "acquisition_date",
    "coordinates",
    "labels",
    "projection",
    "tile_source",
}

BEN_S1_V1_0_TAR_MD5SUM = "94ced73440dea8c7b9645ee738c5a172"
BEN_S2_V1_0_TAR_MD5SUM = "5a64e9ce38deb036a435a7b59494924c"


class SentinelSource(str, Enum):
    """
    BigEarthNet Sentinel-Source S1 or S2.
    """

    S1 = "S1"
    S2 = "S2"

    def __str__(self):
        return self.value


class Season(str, Enum):
    """
    Seasons of the year.
    """

    Winter = "Winter"
    Fall = "Fall"
    Summer = "Summer"
    Spring = "Spring"

    def __str__(self):
        return self.value


_COUNTRY_TO_ISO_A2 = {
    "Austria": "AT",
    "Belgium": "BE",
    "Finland": "FI",
    "Ireland": "IE",
    "Kosovo": "XK",
    "Lithuania": "LT",
    "Luxembourg": "LU",
    "Portugal": "PT",
    "Serbia": "RS",
    "Switzerland": "CH",
}


class Country(str, Enum):
    """
    Official BigEarthNet countries
    """

    Austria = "Austria"
    Belgium = "Belgium"
    Finland = "Finland"
    Ireland = "Ireland"
    Kosovo = "Kosovo"
    Lithuania = "Lithuania"
    Luxembourg = "Luxembourg"
    Portugal = "Portugal"
    Serbia = "Serbia"
    Switzerland = "Switzerland"

    def to_iso_A2(self):
        return _COUNTRY_TO_ISO_A2[self.value]

    def __str__(self):
        return self.value


class Split(str, Enum):
    """
    Official split _names_ to keep the naming consistent.
    """

    train = "train"
    validation = "validation"
    test = "test"

    def __str__(self):
        return self.value


COUNTRIES = tuple(c for c in Country)

# NOTE: ISO_A2 because ISO_A3 does is NOT defined for Kosovo!
COUNTRIES_ISO_A2 = tuple(c.to_iso_A2() for c in Country)


BEN_COMPLETE_SIZE = 590_326
BEN_SNOWY_PATCHES_COUNT = 61_707
BEN_CLOUDY_OR_SHADOWY_PATCHES_COUNT = 9_280
# this is before removing cloudy/snowy patches
BEN_NO_19_CLASS_TARGET_COUNT = 57
BEN_RECOMMENDED_SIZE = 519_284


# From: https://depositonce.tu-berlin.de/bitstream/11303/11261/3/BigEarthNetManual.pdf
# Ignoring the mistakenly leading 0 in vertical and horizontal patch.
# Ignore the bug where the hour digits do not necessarily have a leading 0

BEN_S2_RE = re.compile(
    r"""
        (?P<sentinel_mission>S2[AB])
        _
        MSIL2A # Sentinel-2 data product
        _
        (?P<year>\d{4})
        (?P<month>\d{2})
        (?P<day>\d{2})
        T
        (?P<hour>\d{1,2}) # Bug
        (?P<minute>\d{2})
        (?P<second>\d{2})
        _
        (?P<horizontal_id>\d{1,2})
        _
        (?P<vertical_id>\d{1,2})
    """,
    re.VERBOSE,
)


# https://sentinels.copernicus.eu/web/sentinel/user-guides/sentinel-1-sar/naming-conventions
BEN_S1_RE = re.compile(
    r"""
        (?P<sentinel_mission>S1[AB])
        _
        IW # Interferometric Wide swath
        _
        GRDH # Ground-Range-Detection High Resolution
        _
        1   # processing level 1
        S   # Product class Standard
        DV  # Polarisation dual VV+VH polarisation
        _
        (?P<year>\d{4})
        (?P<month>\d{2})
        (?P<day>\d{2})
        T
        # Inspired by the bug from Sentinel-2
        # Checked all _current_ Sentinel-1 patches
        # and all of them use two digits for hour
        (?P<hour>\d{1,2})
        (?P<minute>\d{2})
        (?P<second>\d{2})
        _
        (?P<sentinel_2_l1c_tile_area>\w{5})
        _
        (?P<horizontal_id>\d{1,2})
        _
        (?P<vertical_id>\d{1,2})
    """,
    re.VERBOSE,
)


# works because of greedy regex
BEN_S1_BAND_RE = re.compile(r".*_(?P<band>V[HV])")
BEN_S2_BAND_RE = re.compile(r".*(?P<band>B\d[0-9A])")


# patch that lies outside of Finland and is easily missed
# by simply intersection calculations based on country shapefiles
# as it would probably belong to Russia
PATCH_FROM_RUSSIA = "S2B_MSIL2A_20180221T093029_65_1"
# patch that lies in the sea-region of Finland
PATCH_IN_TERROTORIAL_WATERS = "S2B_MSIL2A_20170814T100029_33_77"

from collections.abc import Mapping, Sequence
from functools import reduce, singledispatch

from rich.table import Table


def _single_column_table(col_name, rows):
    t = Table(col_name)
    for row in rows:
        t.add_row(row)
    return t


def _simple_dict_table(header, dictionary):
    t = Table(title=header, show_header=False)
    for k, v in dictionary.items():
        t.add_row(str(k), str(v))
    return t


def _default_pprint(value, name):
    rich.print(f"{name}: ", value)


@singledispatch
def _smart_pprint(value, name):
    _default_pprint(value, name)


@_smart_pprint.register
def _(value: Mapping, name):
    if all(isinstance(v, (str, int, float)) or v is None for v in value.values()):
        t = _simple_dict_table(name, value)
        rich.print(t)
    else:
        _default_pprint(value, name)


@_smart_pprint.register
def _(value: str, name):
    # default_pprint str and not call Sequence branch
    _default_pprint(value, name)


@_smart_pprint.register
def _(value: Sequence, name):
    if all(isinstance(entry, str) for entry in value):
        t = _single_column_table(name, value)
        rich.print(t)
    else:
        _default_pprint(value, name)


def smart_pprint(name, value):
    """
    A small `rich.print` wrapper that tries to guess a good representation
    for variable named `name` with the given `value`.
    If no match is found, the function will simply call `rich.print` under the hood.
    Lists of basic types will be printed as a single column list, whereas simple
    dictionaries will be printed as two-column tables, for example.
    """
    _smart_pprint(value, name)


def print_all_constants():
    """
    A function that shows all of the pre-defined constants of the library.
    """
    ben_constants = {
        k: v for k, v in globals().items() if not k.startswith("_") and k.upper() == k
    }
    for k, v in ben_constants.items():
        smart_pprint(k, v)


import click

# FUTURE: should use typer with option
# as it provides an easy interface to generate autocompletions for options
# with some more advanced features.


def cli():
    """
    A function that returns a `click` based CLI application.
    Should be called from `__main__`.
    """
    # where name is selected by user via an option
    # and the value is retrieved from the dictionary

    ben_constants = {
        k: v for k, v in globals().items() if not k.startswith("_") and k.upper() == k
    }

    def smart_pprint_constant(**kwargs):
        """
        Quickly visualize BigEarthNet constants.
        Select one or more constants to print.
        """
        # value True if flag set; False otherwise
        for name, set in kwargs.items():
            name = name.upper()  # undo the click normalization "lower"
            name = name.replace("-", "_")  # undo the personal replacement
            if set:
                smart_pprint(name.upper(), ben_constants[name])

    print_func = smart_pprint_constant
    for k in ben_constants.keys():
        print_func = click.option(f"--{k.replace('_', '-')}", is_flag=True)(print_func)

    cmd = click.command()(print_func)
    return cmd()


if __name__ == "__main__":
    # constants_prompt()
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bigearthnet_common']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4,<2.0',
 'fastcore>=1.3,<2.0',
 'natsort>=8,<9',
 'pydantic>=1.8,<2.0',
 'python-dateutil>=2,<3',
 'rich-click[typer]>=1.3.2,<2.0.0',
 'rich>=10,<13',
 'typer>=0.4,<0.5']

entry_points = \
{'console_scripts': ['ben_build_csv_sets = '
                     'bigearthnet_common.sets:build_csv_sets_cli',
                     'ben_constant_prompt = bigearthnet_common.constants:cli',
                     'ben_describe_patch = '
                     'bigearthnet_common.base:describe_patch_cli',
                     'ben_validate_s1_root_dir = '
                     'bigearthnet_common.base:validate_ben_s1_root_directory_cli',
                     'ben_validate_s2_root_dir = '
                     'bigearthnet_common.base:validate_ben_s2_root_directory_cli']}

setup_kwargs = {
    'name': 'bigearthnet-common',
    'version': '2.7.0',
    'description': 'A collection of common tools to interact with the BigEarthNet dataset.',
    'long_description': '# BigEarthNet Common\n> A personal collection of common tools to interact with the BigEarthNet dataset.\n\n[![Tests](https://img.shields.io/github/workflow/status/kai-tub/bigearthnet_common/CI?color=dark-green&label=%20Tests)](https://github.com/kai-tub/bigearthnet_common/actions/workflows/main.yml)\n[![Python Versions](https://img.shields.io/pypi/pyversions/bigearthnet_common)](https://pypi.org/project/bigearthnet_common)\n[![License](https://img.shields.io/pypi/l/bigearthnet_common?color=dark-green)](https://github.com/kai-tub/bigearthnet_common/blob/main/LICENSE)\n[![PyPI version](https://badge.fury.io/py/bigearthnet-common.svg)](https://pypi.org/project/bigearthnet-common/)\n[![Conda Version](https://img.shields.io/conda/vn/conda-forge/bigearthnet-common?color=dark-green)](https://anaconda.org/conda-forge/bigearthnet-common)\n[![Auto Release](https://img.shields.io/badge/release-auto.svg?colorA=888888&colorB=9B065A&label=auto&logo=data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABQAAAAUCAYAAACNiR0NAAACzElEQVR4AYXBW2iVBQAA4O+/nLlLO9NM7JSXasko2ASZMaKyhRKEDH2ohxHVWy6EiIiiLOgiZG9CtdgG0VNQoJEXRogVgZYylI1skiKVITPTTtnv3M7+v8UvnG3M+r7APLIRxStn69qzqeBBrMYyBDiL4SD0VeFmRwtrkrI5IjP0F7rjzrSjvbTqwubiLZffySrhRrSghBJa8EBYY0NyLJt8bDBOtzbEY72TldQ1kRm6otana8JK3/kzN/3V/NBPU6HsNnNlZAz/ukOalb0RBJKeQnykd7LiX5Fp/YXuQlfUuhXbg8Di5GL9jbXFq/tLa86PpxPhAPrwCYaiorS8L/uuPJh1hZFbcR8mewrx0d7JShr3F7pNW4vX0GRakKWVk7taDq7uPvFWw8YkMcPVb+vfvfRZ1i7zqFwjtmFouL72y6C/0L0Ie3GvaQXRyYVB3YZNE32/+A/D9bVLcRB3yw3hkRCdaDUtFl6Ykr20aaLvKoqIXUdbMj6GFzAmdxfWx9iIRrkDr1f27cFONGMUo/gRI/jNbIMYxJOoR1cY0OGaVPb5z9mlKbyJP/EsdmIXvsFmM7Ql42nEblX3xI1BbYbTkXCqRnxUbgzPo4T7sQBNeBG7zbAiDI8nWfZDhQWYCG4PFr+HMBQ6l5VPJybeRyJXwsdYJ/cRnlJV0yB4ZlUYtFQIkMZnst8fRrPcKezHCblz2IInMIkPzbbyb9mW42nWInc2xmE0y61AJ06oGsXL5rcOK1UdCbEXiVwNXsEy/6+EbaiVG8eeEAfxvaoSBnCH61uOD7BS1Ul8ESHBKWxCrdyd6EYNKihgEVrwOAbQruoytuBYIFfAc3gVN6iawhjKyNCEpYhVJXgbOzARyaU4hCtYizq5EI1YgiUoIlT1B7ZjByqmRWYbwtdYjoWoN7+LOIQefIqKawLzK6ID69GGpQgwhhEcwGGUzfEPAiPqsCXadFsAAAAASUVORK5CYII=)](https://github.com/intuit/auto)\n\nThis library provides a collection of high-level tools to better work with the [BigEarthNet](www.bigearth.net) dataset.\n\n`bigearthnet_common` tries to:\n\n1. Collect the most relevant _constants_ into a single place to reduce the time spent looking for these, like:\n   - The 19 or 43 class nomenclature strings\n   - URL\n   - Band statistics (mean/variance) as integer and float\n   - Channel names\n   - etc.\n2. Provide common metadata related functions\n   - Safe JSON parser for S1/S2\n   - Get the original split\n   - Get a list of snowy/cloudy patches\n   - Convert the _old_ labels to the _new_ label nomenclature\n   - Generate multi-hot encoded label views\n3. Easily filter patches and generate subsets as CSV files\n4. Allow to quickly test code on BigEarthNet data without requiring to download the entire archvie\n\n## Installation\nThe package is available via PyPI and can be installed with:\n- `pip install bigearthnet_common`\n\nThe package has _Python-only_ dependencies and should cause no issues in more complex Conda environments with various binaries.\n\n## Review constants\nTo quickly search for BigEarthNet constants of interest, call:\n- `ben_constants_prompt` or\n- `python -m bigearthnet_common.constants`\n\n## Sets generator\nTo generate sets/subsets from the data and to store them as a CSV file, use:\n- `ben_build_csv_sets --help`\n\nThis command-line tool lets the user easily create subsets from common constraints.\nTo generate a CSV file that contains all Sentinel-2 patches from Serbia only during the Summer and Spring months, call the function as:\n- `ben_build_csv_sets <FILE_PATH> S2 --seasons Winter --seasons Summer --countries Serbia --remove-unrecommended-dl-patches`\n\n:::{note}\n\nBy default, this tool will always remove the _unrecommended_ patches, i.e., patches that contain seasonal snow, shadows, clouds, or that have no labels in the 19-class nomenclature\n\n:::\n\n\n## Describe patch\nThe library provides a tool to quickly visualize meta-data information about each patch:\n```sh\nben_describe_patch <S1/S2 patch name>\n```\n',
    'author': 'Kai Norman Clasen',
    'author_email': 'k.clasen@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kai-tub/bigearthnet_common',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)
